/*
 * Copyright (c) 2014-2017 Satlab Aps <satlab@satlab.com>
 * All rights reserved. Do not distribute without permission.
 */

#ifndef _SL_TRACE_H_
#define _SL_TRACE_H_

#include <stdbool.h>
#include <stdint.h>
#include <stdarg.h>

#include <util/colorprint.h>

typedef enum {
	TRACE_LEVEL_DEBUG = 0,
	TRACE_LEVEL_INFO,
	TRACE_LEVEL_WARN,
	TRACE_LEVEL_ERROR,
	/* Must be last */
	__TRACE_LEVEL_MAX,
} trace_level_t;

#define TRACE_MASK_DEBUG (1 << TRACE_LEVEL_DEBUG)
#define TRACE_MASK_INFO  (1 << TRACE_LEVEL_INFO)
#define TRACE_MASK_WARN	 (1 << TRACE_LEVEL_WARN)
#define TRACE_MASK_ERROR (1 << TRACE_LEVEL_ERROR)
#define TRACE_MASK_ALL \
	(TRACE_MASK_ERROR | TRACE_MASK_WARN | TRACE_MASK_INFO | TRACE_MASK_DEBUG)
#define TRACE_MASK_DEFAULT \
	(TRACE_MASK_ERROR | TRACE_MASK_WARN)

#define TRACE_BUFFER_SIZE	2048
#define TRACE_MESSAGE_SIZE	256

struct trace_group {
	const char *name;
	trace_level_t print_mask;
	trace_level_t store_mask;
};

struct trace_level {
	const char *name;
	sl_color_t color;
	unsigned int mask;
};

bool trace_level_enabled(trace_level_t level, struct trace_group *const group);

int trace_message(trace_level_t level, struct trace_group *group, const char *format, ...) __attribute__ ((format (printf, 3, 4)));

int trace_message_va(trace_level_t level, struct trace_group *group, const char *format, va_list args);

extern struct trace_group *const TRACE_DEFAULT;

#define trace_message_once(_level, _group, _fmt, ...)	\
({						\
	static bool print_once;			\
	if (!print_once) {			\
		print_once = true;		\
		trace_message(_level, _group, _fmt, ##__VA_ARGS__); \
	}					\
})

#define trace_message_ratelimit(_level, _group, _ms, _fmt, ...)	\
({								\
 	if (trace_level_enabled(_level, _group)) {		\
		static uint64_t last = 0;			\
		uint64_t now = timer_getmonotonic();		\
		if (now > last + (uint64_t)_ms * 1000000ULL) {	\
			last = now;				\
			trace_message(_level, _group, _fmt, ##__VA_ARGS__); \
		}						\
	}							\
})

#define trace_message_on(_level, _group, _cond, _fmt, ...)	\
({								\
 	if (_cond) {						\
		trace_message(_level, _group, _fmt, ##__VA_ARGS__); \
	}							\
})

#define trace_error_enabled() \
	trace_level_enabled(TRACE_LEVEL_ERROR, TRACE_DEFAULT)
#define trace_warning_enabled() \
	trace_level_enabled(TRACE_LEVEL_WARN, TRACE_DEFAULT)
#define trace_info_enabled() \
	trace_level_enabled(TRACE_LEVEL_INFO, TRACE_DEFAULT)
#define trace_debug_enabled() \
	trace_level_enabled(TRACE_LEVEL_DEBUG, TRACE_DEFAULT)

#define pr_error(_fmt, ...) \
	trace_message(TRACE_LEVEL_ERROR, TRACE_DEFAULT, _fmt, ##__VA_ARGS__)
#define pr_warning(_fmt, ...) \
	trace_message(TRACE_LEVEL_WARN, TRACE_DEFAULT, _fmt, ##__VA_ARGS__)
#define pr_info(_fmt, ...) \
	trace_message(TRACE_LEVEL_INFO, TRACE_DEFAULT, _fmt, ##__VA_ARGS__)
#define pr_debug(_fmt, ...) \
	trace_message(TRACE_LEVEL_DEBUG, TRACE_DEFAULT, _fmt, ##__VA_ARGS__)

#define pr_error_once(_fmt, ...) \
	trace_message_once(TRACE_LEVEL_ERROR, TRACE_DEFAULT, _fmt, ##__VA_ARGS__)
#define pr_warning_once(_fmt, ...) \
	trace_message_once(TRACE_LEVEL_WARN, TRACE_DEFAULT, _fmt, ##__VA_ARGS__)
#define pr_info_once(_fmt, ...) \
	trace_message_once(TRACE_LEVEL_INFO, TRACE_DEFAULT, _fmt, ##__VA_ARGS__)
#define pr_debug_once(_fmt, ...) \
	trace_message_once(TRACE_LEVEL_DEBUG, TRACE_DEFAULT, _fmt, ##__VA_ARGS__)

#define pr_error_ratelimit(_ms, _fmt, ...) \
	trace_message_ratelimit(TRACE_LEVEL_ERROR, TRACE_DEFAULT, _ms, _fmt, ##__VA_ARGS__)
#define pr_warning_ratelimit(_ms, _fmt, ...) \
	trace_message_ratelimit(TRACE_LEVEL_WARN, TRACE_DEFAULT, _ms, _fmt, ##__VA_ARGS__)
#define pr_info_ratelimit(_ms, _fmt, ...) \
	trace_message_ratelimit(TRACE_LEVEL_INFO, TRACE_DEFAULT, _ms, _fmt, ##__VA_ARGS__)
#define pr_debug_ratelimit(_ms, _fmt, ...) \
	trace_message_ratelimit(TRACE_LEVEL_DEBUG, TRACE_DEFAULT, _ms, _fmt, ##__VA_ARGS__)

#define pr_error_on(_cond, _fmt, ...) \
	trace_message_on(TRACE_LEVEL_ERROR, TRACE_DEFAULT, _cond, _fmt, ##__VA_ARGS__)
#define pr_warning_on(_cond, _fmt, ...) \
	trace_message_on(TRACE_LEVEL_WARN, TRACE_DEFAULT, _cond, _fmt, ##__VA_ARGS__)
#define pr_info_on(_cond, _fmt, ...) \
	trace_message_on(TRACE_LEVEL_INFO, TRACE_DEFAULT, _cond, _fmt, ##__VA_ARGS__)
#define pr_debug_on(_cond, _fmt, ...) \
	trace_message_on(TRACE_LEVEL_DEBUG, TRACE_DEFAULT, _cond, _fmt, ##__VA_ARGS__)

#define TRACE_GROUP_MASKED(_group, _name, _mask)	\
	__attribute__((section("trace_groups"),used))	\
	struct trace_group _group ## _s = { 		\
		.name = _name,				\
		.print_mask = _mask,			\
		.store_mask = _mask,			\
	};						\
	struct trace_group *const _group = &_group ## _s

#define TRACE_GROUP(_group, _name)			\
	TRACE_GROUP_MASKED(_group, _name, TRACE_MASK_DEFAULT)

#define TRACE_GROUP_EXTERN(_group)			\
	extern struct trace_group *const _group

int trace_mask_from_string(char *str, trace_level_t *current);

int trace_groups_iterate(void (trace_group_iter)(struct trace_group *));

struct trace_group *trace_group_find(const char *groupstr);

int trace_level_from_string(const char *levelstr);

#endif /* _SL_TRACE_H_ */
