/*
 * Copyright (c) 2014-2017 Satlab Aps <satlab@satlab.com>
 * All rights reserved. Do not distribute without permission.
 */

#include <util/trace.h>

#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>

#include <util/util.h>
#include <util/colorprint.h>
#include <util/timer.h>
#include <slash/slash.h>

/* Created by linker */
extern struct trace_group __start_trace_groups;
extern struct trace_group __stop_trace_groups;

static struct trace_level levels[] = {
	[TRACE_LEVEL_DEBUG] = {
		.name = "debug",
		.color =  COLOR_DARK_GRAY,
		.mask = TRACE_MASK_DEBUG},
	[TRACE_LEVEL_INFO]  = {
		.name = "info",
		.color = COLOR_GREEN,
		.mask = TRACE_MASK_INFO},
	[TRACE_LEVEL_WARN]  = {
		.name = "warn",
		.color = COLOR_YELLOW,
		.mask = TRACE_MASK_WARN},
	[TRACE_LEVEL_ERROR] = {
		.name = "error",
		.color = COLOR_RED,
		.mask = TRACE_MASK_ERROR},
};

/* Message circular buffer */
static char trace_message_buf[TRACE_BUFFER_SIZE];

/* Default catch-all trace group */
TRACE_GROUP(tg_default, "default");
struct trace_group *const TRACE_DEFAULT = &tg_default_s;

bool trace_level_enabled(trace_level_t level, struct trace_group *const group)
{
	trace_level_t mask;

	/* Skip if level is invalid */
	if (level >= __TRACE_LEVEL_MAX)
		return false;

	/* Skip if level not enabled */
	mask = levels[level].mask;
	if (!(mask & (group->print_mask | group->store_mask)))
		return false;

	return true;
}

int trace_message_va(trace_level_t level,
		     struct trace_group *const group,
		     const char *format, va_list args)
{
	int n = 0;
	uint64_t now;
	uint32_t sec, nsec;
	char buf[TRACE_MESSAGE_SIZE];
	trace_level_t mask;

	/* Skip if level is invalid or disabled */
	if (!trace_level_enabled(level, group))
		return 0;

	now = timer_getmonotonic();
	sec = NSEC_TO_SEC(now);
	nsec = now - sec * NSEC_PER_SEC;

	n += color_set(buf + n, COLOR_LIGHT_GRAY);
	n += snprintf(buf + n, TRACE_MESSAGE_SIZE - n,
		      "[%5u.%06u] ",
		      (unsigned int)sec,
		      (unsigned int)nsec/1000U);
	n += color_set(buf + n, levels[level].color);
	n += snprintf(buf + n, TRACE_MESSAGE_SIZE - n,
		      "%s: ", group->name);
	n += color_set(buf + n, COLOR_RESET);
	n += vsnprintf(buf + n, TRACE_MESSAGE_SIZE - n - 1, format, args);

	/* Do not keep linefeed */
	if (buf[n - 1] == '\n')
		buf[n - 1] = '\0';

	mask = levels[level].mask;
	if (mask & group->store_mask) {
		/* FIXME: put in circular buffer */
		(void)trace_message_buf;
	}

	if (mask & group->print_mask) {
		printf("%s\n", buf);
		fflush(stdout);
	}

	return n;
}

int trace_message(trace_level_t level, struct trace_group *group,
		  const char *format, ...)
{
	int n;
	va_list args;

	va_start(args, format);

	n = trace_message_va(level, group, format, args);

	va_end(args);

	return n;
}

#define TRACE_MASK_LEVEL 0
#define TRACE_MASK_CLEAR 1
#define TRACE_MASK_SET   2

int trace_mask_from_string(char *str, trace_level_t *current)
{
	int op;
	char *token, *saveptr;
	trace_level_t mask;

	token = strtok_r(str, ",", &saveptr);
	while (token != NULL) {
		if (*token == '+') {
			op = TRACE_MASK_SET;
			token++;
		} else if (*token == '-') {
			op = TRACE_MASK_CLEAR;
			token++;
		} else {
			op = TRACE_MASK_LEVEL;
		}

		if (strlen(token) < 1)
			return -1;

		if (!strncasecmp(token, "debug", strlen(token))) {
			mask = TRACE_MASK_DEBUG;
		} else if (!strncasecmp(token, "info", strlen(token))) {
			mask = TRACE_MASK_INFO;
		} else if (!strncasecmp(token, "warning", strlen(token))) {
			mask = TRACE_MASK_WARN;
		} else if (!strncasecmp(token, "error", strlen(token))) {
			mask = TRACE_MASK_ERROR;
		} else if (!strncasecmp(token, "default", strlen(token))) {
			mask = TRACE_MASK_DEFAULT;
			op = TRACE_MASK_LEVEL;
		} else if (!strncasecmp(token, "all", strlen(token))) {
			mask = TRACE_MASK_ALL;
			op = TRACE_MASK_LEVEL;
		} else if (!strncasecmp(token, "none", strlen(token))) {
			mask = 0;
			op = TRACE_MASK_LEVEL;
		} else {
			return -EINVAL;
		}

		if (op == TRACE_MASK_SET) {
			*current |= mask;
		} else if (op == TRACE_MASK_CLEAR) {
			*current &= ~mask;
		} else if (op == TRACE_MASK_LEVEL) {
			*current = mask | ~(mask - 1);
		}

		token = strtok_r(NULL, ",", &saveptr);
	}

	return 0;
}

int trace_groups_iterate(void (trace_group_iter)(struct trace_group *))
{
	struct trace_group *group;

	for (group = &__start_trace_groups;
	     group < &__stop_trace_groups;
	     group++) {
		trace_group_iter(group);
	}

	return 0;
}

struct trace_group *trace_group_find(const char *groupstr)
{
	struct trace_group *group = NULL;

	for (group = &__start_trace_groups;
	     group < &__stop_trace_groups;
	     group++) {
		if (!strcmp(groupstr, group->name))
			return group;
	}

	return NULL;
}

int trace_level_from_string(const char *levelstr)
{
	int i, level = -EINVAL;

	for (i = 0; i < ARRAY_SIZE(levels); i++) {
		if (!strcmp(levelstr, levels[i].name)) {
			level = i;
			break;
		}
	}

	return level;
}
